import json
from urllib.parse import urlparse

import requests
from jsonschema import FormatChecker
from jsonschema.validators import Draft4Validator as validator

from .base import BaseCommand
from ocdskit.exceptions import CommandError


class Command(BaseCommand):
    name = 'validate'
    help = 'reads JSON data from standard input, validates it against the schema, and prints errors'

    def add_arguments(self):
        self.add_argument('--schema', help='the schema to validate against',
                          default='http://standard.open-contracting.org/latest/en/release-package-schema.json')

    def handle(self):
        components = urlparse(self.args.schema)
        if components.scheme == 'file':
            with open(self.args.schema[7:]) as f:
                schema = json.loads(f.read())
        else:
            schema = requests.get(self.args.schema).json()

        for i, line in enumerate(self.buffer()):
            try:
                data = json.loads(line)
                for error in validator(schema, format_checker=FormatChecker()).iter_errors(data):
                    print('item {}: {} ({})'.format(i, error.message, '/'.join(error.absolute_schema_path)))
            except json.decoder.JSONDecodeError as e:
                raise CommandError('item {}: JSON error: {}'.format(i, e))



import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

import pytest

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('realdata/release-package-1.json', 'rb') + read('realdata/release-package-2.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'compile'])
        main()

    assert actual.getvalue() == read('realdata/compiled-release-1.json') + read('realdata/compiled-release-2.json')


def test_command_help(monkeypatch, caplog):
    stdin = read('release-package_minimal.json', 'rb')

    with pytest.raises(SystemExit) as excinfo:
        with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
            monkeypatch.setattr(sys, 'argv', ['ocdskit', '--help'])
            main()

    assert len(caplog.records()) == 0
    assert excinfo.value.code == 0

    assert actual.getvalue().startswith('usage: ocdskit [-h] ')


def test_command_pretty(monkeypatch):
    stdin = read('release-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', '--pretty', 'compile'])
        main()

    assert actual.getvalue() == read('compile_pretty_minimal.json')


def test_command_encoding(monkeypatch, caplog):
    stdin = read('realdata/release-package_encoding-iso-8859-1.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', '--encoding', 'iso-8859-1', 'compile'])
        main()

    assert actual.getvalue() == read('realdata/compile_encoding_encoding.json')


def test_command_bad_encoding_iso_8859_1(monkeypatch, caplog):
    stdin = read('realdata/release-package_encoding-iso-8859-1.json', 'rb')

    with pytest.raises(SystemExit) as excinfo:
        with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO):
            monkeypatch.setattr(sys, 'argv', ['ocdskit', 'compile'])
            main()

    assert len(caplog.records()) == 1
    assert caplog.records()[0].levelname == 'CRITICAL'
    assert caplog.records()[0].message == "encoding error: try `--encoding iso-8859-1`? ('utf-8' codec can't decode " \
                                          "byte 0xd3 in position 592: invalid continuation byte)"
    assert excinfo.value.code == 1



import json
import io
import sys


class BaseCommand(object):
    def __init__(self, subparsers):
        """
        Initializes the subparser and adds arguments.
        """
        self.subparser = subparsers.add_parser(self.name, description=self.help)
        self.add_arguments()

    def add_arguments(self):
        pass

    def add_argument(self, *args, **kwargs):
        """
        Adds an argument to the subparser.
        """
        self.subparser.add_argument(*args, **kwargs)

    def handle(self):
        raise NotImplementedError('commands must implement handle()')

    def buffer(self):
        return io.TextIOWrapper(sys.stdin.buffer, encoding=self.args.encoding)

    def print(self, data):
        """
        Prints JSON data.
        """
        # See https://docs.python.org/2/library/json.html
        if self.args.pretty:
            kwargs = {'indent': 2, 'separators': (',', ': ')}
        else:
            kwargs = {'separators': (',', ':')}

        print(json.dumps(data, **kwargs))



import json
from collections import OrderedDict

from .base import BaseCommand


class Command(BaseCommand):
    name = 'combine-release-packages'
    help = 'reads release packages from standard input, collects releases, and prints one release package'

    def handle(self):
        output = OrderedDict([('extensions', OrderedDict()), ('releases', [])])

        for line in self.buffer():
            package = json.loads(line, object_pairs_hook=OrderedDict)

            # Use sample metadata.
            output['uri'] = package['uri']
            output['publishedDate'] = package['publishedDate']
            output['publisher'] = package['publisher']

            if 'extensions' in package:
                # Python has no OrderedSet, so we use OrderedDict to keep extensions in order without duplication.
                output['extensions'].update(dict.fromkeys(package['extensions'], True))

            for field in ('license', 'publicationPolicy', 'version'):
                if field in package:
                    output[field] = package[field]

            output['releases'].extend(package['releases'])

        if output['extensions']:
            output['extensions'] = list(output['extensions'].keys())
        else:
            del output['extensions']

        self.print(output)



import json

import jsonref
import sqlalchemy
from sqlalchemy.dialects.postgresql import JSONB

from .base import BaseCommand


class Command(BaseCommand):
    name = 'tabulate'
    help = 'load packages into a database'

    def add_arguments(self):
        self.add_argument('database_url', help='sqlalchemy database URL')
        self.add_argument('--drop', help='drop all tables before loading', action='store_true')
        self.add_argument('--schema', help='the release-schema.json to use',
                          default='http://standard.open-contracting.org/latest/en/release-schema.json')

    def handle(self):
        deref_schema = jsonref.load_uri(self.args.schema)
        metadata, engine = self.create_db(self.args.database_url, deref_schema, drop=self.args.drop)
        self.upload_files(metadata, engine, deref_schema)

    def process_schema_object(self, path, current_name, flattened, obj):
        """
        Return a dictionary with a flattened representation of the schema. `patternProperties` are skipped as we don't
        want them as field names (a regular expression string) in the database.
        """
        properties = obj.get('properties', {})  # an object may have patternProperties only
        current_object = flattened.get(path)

        if current_object is None:
            current_object = {}
            flattened[path] = current_object

        for name, prop in list(properties.items()):
            prop_type = prop['type']
            if prop_type == 'object':
                flattened = self.process_schema_object(path, current_name + name + '_', flattened, prop)
            elif prop_type == 'array':
                if 'object' not in prop['items']['type']:
                    current_object[current_name + name] = prop['items']['type'] + '[]'
                else:
                    flattened = self.process_schema_object(path + (current_name + name,), '', flattened, prop['items'])
            else:
                current_object[current_name + name] = prop_type

        return flattened

    def create_db(self, sqlalchemy_url, deref_schema, drop=False):
        engine = sqlalchemy.create_engine(sqlalchemy_url)
        metadata = sqlalchemy.MetaData(engine)

        if drop:
            metadata.reflect()
            for table in reversed(metadata.sorted_tables):
                table.drop()
            metadata = sqlalchemy.MetaData(engine)

        flattened = {}
        flat = self.process_schema_object((), '', flattened, deref_schema)

        for table, fields in flat.items():
            table_name = '_'.join(table)
            if not table_name:
                table_name = 'releases'
            columns = [sqlalchemy.Column('ocid', sqlalchemy.Text)]
            for parent_table in ('releases',) + table:
                columns.append(sqlalchemy.Column(parent_table[:-1] + '_id', sqlalchemy.Text))

            for field in sorted(fields.keys()):
                types = fields[field]
                if field == 'id':
                    continue
                if 'string' in types:
                    columns.append(sqlalchemy.Column(field, sqlalchemy.Text))
                elif 'number' in types:
                    columns.append(sqlalchemy.Column(field, sqlalchemy.Numeric))
                elif 'integer' in types:
                    columns.append(sqlalchemy.Column(field, sqlalchemy.Integer))
                else:
                    columns.append(sqlalchemy.Column(field, sqlalchemy.Text))
            if 'postgresql' in engine.name:
                columns.append(sqlalchemy.Column('extras', JSONB))
            else:
                columns.append(sqlalchemy.Column('extras', sqlalchemy.Text))

            sqlalchemy.Table(table_name, metadata, *columns)

        metadata.create_all(engine)
        return metadata, engine

    def process_object(self, path, current_name, current_keys, flattened, obj, flat_obj):
        if flat_obj is None:
            flat_list = flattened.get(path)
            if flat_list is None:
                flat_list = []
                flattened[path] = flat_list
            new_id = obj.pop('id', None)
            if not new_id:
                new_id = str(len(flat_list) + 1)
            if not current_keys:
                current_keys = (obj['ocid'],)
            current_keys = current_keys + (new_id,)
            flat_obj = {'ocid': current_keys[0]}
            for num, table in enumerate(('releases',) + path):
                flat_obj[table[:-1] + '_id'] = current_keys[num + 1]
            flat_list.append(flat_obj)

        for key, value in obj.items():
            if isinstance(value, dict):
                self.process_object(path, current_name + key + '_', current_keys, flattened, value, flat_obj)
            elif isinstance(value, list):
                if not value:
                    continue
                if isinstance(value[0], dict):
                    for item in value:
                        self.process_object(path + (current_name + key,), '', current_keys, flattened, item, None)
                else:
                    flat_obj[current_name + key] = json.dumps(value)
            else:
                flat_obj[current_name + key] = value
        return flattened

    def upload_files(self, metadata, engine, deref_schema):
        for line in self.buffer():
            package = json.loads(line)
            if 'releases' in package:
                releases = package['releases']
            elif 'records' in package:
                releases = []
                for record in package['records']:
                    releases.extend(record['releases'])

            self.upload_file(metadata, engine, deref_schema, releases)

    def upload_file(self, metadata, engine, deref_schema, releases):
        conn = engine.connect()

        tabulated = {}
        for release in releases:
            tabulated = self.process_object(tuple(), '', tuple(), tabulated, release, None)

        for table, rows in tabulated.items():
            if not table:
                table_name = 'releases'
            else:
                table_name = '_'.join(table)
            try:
                table = metadata.tables[table_name]
            except KeyError:
                print('table {} does not exist'.format(table_name))
                continue

            for row in rows:
                for key in table.c.keys():
                    if key not in row.keys():
                        row[key] = None
                extras = {}
                for key in list(row.keys()):
                    if key not in table.c:
                        extras[key] = row.pop(key)
                if 'postgresql' in engine.name:
                    row['extras'] = extras
                else:
                    row['extras'] = json.dumps(extras)
            conn.execute(table.insert(), rows)



import os
import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

import pytest

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('realdata/release-package-1.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'validate'])
        main()

    assert actual.getvalue() == "item 0: 'version' is a required property (required)\n"


def test_command_invalid_json(monkeypatch, caplog):
    with caplog.atLevel('INFO'):
        stdin = b'{\n'

        with pytest.raises(SystemExit) as excinfo:
            with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO):
                monkeypatch.setattr(sys, 'argv', ['ocdskit', 'validate'])
                main()

        assert len(caplog.records()) == 1
        assert caplog.records()[0].levelname == 'CRITICAL'
        assert caplog.records()[0].message == "item 0: JSON error: Expecting property name enclosed in double " \
                                              "quotes: line 2 column 1 (char 2)"
        assert excinfo.value.code == 1


def test_command_valid_release_package_url(monkeypatch):
    url = 'http://standard.open-contracting.org/schema/1__0__3/release-package-schema.json'

    stdin = read('realdata/release-package-1.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'validate', '--schema', url])
        main()

    assert actual.getvalue() == ''


def test_command_valid_release_package_file(monkeypatch):
    url = 'file://{}'.format(os.path.realpath(os.path.join('tests', 'fixtures', 'release-package-schema.json')))

    stdin = read('realdata/release-package-1.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'validate', '--schema', url])
        main()

    assert actual.getvalue() == ''


def test_command_invalid_record_package(monkeypatch):
    url = 'http://standard.open-contracting.org/latest/en/record-package-schema.json'

    stdin = read('realdata/record-package-1.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'validate', '--schema', url])
        main()

    assert "item 0: None is not of type 'string' (properties/records/items/properties/compiledRelease/properties/tender/properties/submissionMethod/items/type)" in actual.getvalue()  # noqa



import json
from collections import OrderedDict

from .base import BaseCommand


class Command(BaseCommand):
    name = 'combine-record-packages'
    help = 'reads record packages from standard input, collects packages and records, and prints one record package'

    def handle(self):
        output = OrderedDict([('extensions', OrderedDict()), ('packages', []), ('records', [])])

        for line in self.buffer():
            package = json.loads(line, object_pairs_hook=OrderedDict)

            # Use sample metadata.
            output['uri'] = package['uri']
            output['publishedDate'] = package['publishedDate']
            output['publisher'] = package['publisher']

            if 'extensions' in package:
                # Python has no OrderedSet, so we use OrderedDict to keep extensions in order without duplication.
                output['extensions'].update(dict.fromkeys(package['extensions'], True))

            for field in ('license', 'publicationPolicy', 'version'):
                if field in package:
                    output[field] = package[field]

            for field in ('packages', 'records'):
                if field in package:
                    output[field].extend(package[field])

        if not output['packages']:
            del output['packages']

        if output['extensions']:
            output['extensions'] = list(output['extensions'].keys())
        else:
            del output['extensions']

        self.print(output)



import copy
import csv
import json
import re
import sys
from collections import OrderedDict

from jsonref import JsonRef

from .base import BaseCommand


class Command(BaseCommand):
    name = 'mapping-sheet'
    help = 'generates a spreadsheet with all field paths from an OCDS schema'

    def handle(self):
        release = json.loads(self.buffer().read(), object_pairs_hook=OrderedDict)

        release = JsonRef.replace_refs(release)

        # See https://stackoverflow.com/questions/30734682/extracting-url-and-anchor-text-from-markdown-using-python
        INLINE_LINK_RE = re.compile(r'\[([^\]]+)\]\(([^)]+)\)')

        def find_md_links(md):
            return dict(INLINE_LINK_RE.findall(md))

        def remove_links(text, links):
            for key, link in links.items():
                text = text.replace('[' + key + '](' + link + ')', key)
            return text

        def display_links(links):
            return ', '.join(links.values())

        def display_properties(schema, path='', section='', deprecated=''):
            # Create a copy of obj, because there may be references to it from
            # elsewhere in the JSON schema, and we don't want to mutate it in
            # all those places
            obj = copy.deepcopy(schema['properties'])
            required_fields = schema['required'] if 'required' in schema else []
            rows = []
            for field in obj:
                row = {'path': path + field, 'deprecated': deprecated}

                section = row['path'].split('/')[0] if '/' in row['path'] else ''

                row['section'] = section

                # If there was a reference here, prefer the values from that
                if hasattr(obj[field], '__reference__'):
                    obj[field].update(obj[field].__reference__)

                row['title'] = obj[field]['title'] if 'title' in obj[field] else field + '*'

                if 'description' in obj[field]:
                    links = find_md_links(obj[field]['description'])
                    row['description'] = remove_links(obj[field]['description'], links)
                    row['links'] = display_links(links)

                # Type
                if 'type' in obj[field]:
                    # ToDo: Add checking of the required array also.
                    # This checks whether this field is **implicity required**
                    if 'null' in obj[field]['type']:
                        obj[field]['type'].remove('null')
                        required = False
                    else:
                        required = 'string' in obj[field]['type'] or 'integer' in obj[field]['type']

                    if type(obj[field]['type']) in (tuple, list):
                        row['type'] = ', '.join(obj[field]['type'])
                    else:
                        row['type'] = obj[field]['type']
                else:
                    row['type'] = 'unknown'

                # Required field
                if field in required_fields:
                    required = True

                maxn = 'n' if row['type'] == 'array' else '1'
                minn = '1' if required else '0'
                row['range'] = minn + '..' + maxn

                # Format or restrictions
                if 'format' in obj[field]:
                    row['values'] = obj[field]['format']
                elif 'enum' in obj[field]:
                    if None in obj[field]['enum']:
                        obj[field]['enum'].remove(None)
                    row['values'] = 'Codelist: ' + ', '.join(obj[field]['enum'])
                else:
                    row['values'] = ''

                # Check for deprecation
                if 'deprecated' in obj[field]:
                    row['deprecated'] = obj[field]['deprecated'].get('deprecatedVersion', '')
                    row['deprecationNotes'] = obj[field]['deprecated'].get('description', '')

                rows.append(row)

                if 'properties' in obj[field]:
                    rows = rows + display_properties(obj[field], path + field + '/', section, row['deprecated'])

                if 'items' in obj[field]:
                    if 'properties' in obj[field]['items']:
                        if 'title' in obj[field]['items']:
                            rows.append({'section': section, 'path': path + field,
                                         'title': obj[field]['items']['title'],
                                         'description': obj[field]['items'].get('description', ''),
                                         'type': obj[field]['items']['type']})
                        else:
                            pass

                        rows = rows + display_properties(obj[field]['items'], path + field + '/', section,
                                                         row['deprecated'])

            return rows

        schema = display_properties(release)

        w = csv.DictWriter(sys.stdout, ['section', 'path', 'title', 'description', 'type', 'range', 'values', 'links',
                                        'deprecated', 'deprecationNotes'])
        w.writeheader()
        w.writerows(schema)



import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('release-schema.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'mapping-sheet'])
        main()

    assert actual.getvalue() == read('mapping-sheet.csv').replace('\n', '\r\n')  # not sure why



from setuptools import setup, find_packages

setup(
    name='ocdskit',
    version='0.0.1',
    author='James McKinney',
    author_email='james@slashpoundbang.com',
    url='https://github.com/open-contracting/ocdskit',
    description='A suite of command-line tools for working with OCDS',
    platforms=['any'],
    license='BSD',
    packages=find_packages(),
    entry_points='''[console_scripts]
ocdskit = ocdskit.cli.__main__:main''',
    install_requires=[
        'jsonref',
        'jsonschema',
        'ocdsmerge',
        'python-dateutil',
        'pytz',
        'requests',
        'rfc3987',
        'sqlalchemy',
        'strict-rfc3339',
    ],
    extras_require={
        'test': [
            'pytest<3',
            'pytest-capturelog',
            'pytest-cov',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
    ],
)



import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import read


def test_command(monkeypatch):
    stdin = read('record-package_minimal.json', 'rb') + \
            read('record-package_maximal.json', 'rb') + \
            read('record-package_extensions.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-record-packages'])
        main()

    assert actual.getvalue() == read('combine-record-packages_minimal-maximal-extensions.json')


def test_command_no_extensions(monkeypatch):
    stdin = read('record-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))), patch('sys.stdout', new_callable=StringIO) as actual:
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-record-packages'])
        main()

    assert actual.getvalue() == read('combine-record-packages_minimal.json')



"""
Assumptions:
- All awards/suppliers/identifier/id appear in tender/tenderers/identifier/id
"""

import json

from collections import defaultdict, OrderedDict
from datetime import datetime, timedelta
from statistics import median

import ocdsmerge
import pytz
from dateutil import parser

from .base import BaseCommand


class Command(BaseCommand):
    name = 'measure'
    help = 'measure indicators in OCDS data'

    def add_arguments(self):
        self.add_argument('--currency', help='the expected currency')

    def handle(self):
        data = json.loads(self.buffer().read())

        compiled_releases_by_buyer = defaultdict(list)
        min_date_by_buyer = {}

        # Global indicators.

        b2_procurement_method_count = defaultdict(int)
        b3_procurement_method_amount = defaultdict(int)
        b9_number_of_tenderers = OrderedDict([
            ('by `tender/numberOfTenderers`', []),
            ('by unique `tender/tenderers/identifier/id`', []),
        ])

        b3_no_data = []
        b9_no_data = 0

        for release_package in data:
            releases = release_package['releases']
            compiled_release = ocdsmerge.merge(releases)

            try:
                # Read the data.

                ocid = compiled_release['ocid']
                # We use buyer names as keys, because buyer IDs are reused. We can make this optional.
                buyer = str(compiled_release['buyer']['name'])
                tender = compiled_release['tender']
                procurementMethod = tender['procurementMethod']
                numberOfTenderers = tender['numberOfTenderers']
                tenderers = tender['tenderers']
                amount = tender['value']['amount']
                currency = tender['value']['currency']

                # Prepare data for per-buyer indicators.

                compiled_releases_by_buyer[buyer].append(compiled_release)
                compiled_release['date'] = min(release['date'] for release in releases)
                if buyer not in min_date_by_buyer or compiled_release['date'] < min_date_by_buyer[buyer]:
                    min_date_by_buyer[buyer] = compiled_release['date']

                # Calculate global indicators.

                b2_procurement_method_count[str(procurementMethod)] += 1  # if null

                if procurementMethod and amount:
                    if self.args.currency:
                        # Ensure we sum a single currency.
                        assert currency == self.args.currency, '{} uses {}'.format(ocid, currency)
                    b3_procurement_method_amount[procurementMethod] += amount
                else:
                    b3_no_data.append(amount)

                if numberOfTenderers or tenderers:
                    # Perform basic data validation.
                    assert numberOfTenderers == len(tenderers), ocid
                    b9_number_of_tenderers['by `tender/numberOfTenderers`'].append(numberOfTenderers)
                    # There may be double counting of tenderers.
                    value = len(set([tenderer['identifier']['id'] for tenderer in tenderers]))
                    b9_number_of_tenderers['by unique `tender/tenderers/identifier/id`'].append(value)
                else:
                    b9_no_data += 1
            except KeyError as e:
                raise Exception('{}: {}'.format(e, ocid))

        # Per-buyer indicators.

        b0_compiled_releases = {}
        b1_tenderers = {}
        b4_suppliers_count = {}
        b5_suppliers_amount = {}
        b5_other_currencies = defaultdict(int)
        b8_tenderers_to_suppliers = {}
        b8_tenderers_never_suppliers = {}

        today = datetime.now(pytz.utc)
        threshold = today - timedelta(days=180)  # 6 months

        for buyer, compiled_releases in compiled_releases_by_buyer.items():
            min_date = parser.parse(min_date_by_buyer[buyer])
            if len(compiled_releases) >= 10 and min_date <= threshold:
                # Determine the before/after periods.

                if min_date + timedelta(days=730) <= today:  # 2 years
                    pivot = today - timedelta(days=365)  # 1 year
                else:
                    pivot = today - (today - min_date) / 2

                b0_compiled_releases_all = 0
                b0_compiled_releases_new = 0
                b1_tenderers_before = set()
                b1_tenderers_after = set()
                b4_suppliers_count_before = set()
                b4_suppliers_count_after = set()
                b8_tenderers_count = 0
                b8_suppliers_count = 0

                # Calculate per-buyer indicators.

                for compiled_release in compiled_releases:
                    tender = compiled_release['tender']
                    awards = compiled_release['awards']
                    tenderers = tender['tenderers']

                    b1 = [tenderer['identifier']['id'] for tenderer in tenderers]
                    b4 = [supplier['identifier']['id'] for award in awards for supplier in award['suppliers']]
                    b8_tenderers_count += len(b1)
                    b8_suppliers_count += len(b4)

                    b0_compiled_releases_all += 1
                    if parser.parse(compiled_release['date']) < pivot:
                        b1_tenderers_before.update(b1)
                        b4_suppliers_count_before.update(b4)
                    else:
                        b1_tenderers_after.update(b1)
                        b4_suppliers_count_after.update(b4)
                        b0_compiled_releases_new += 1

                b0_compiled_releases[buyer] = b0_compiled_releases_new / b0_compiled_releases_all

                b1_tenderers_all = b1_tenderers_before | b1_tenderers_after
                if b1_tenderers_all:
                    b1_tenderers_new = b1_tenderers_after - b1_tenderers_before
                    b1_tenderers[buyer] = len(b1_tenderers_new) / len(b1_tenderers_all)

                b4_suppliers_count_all = b4_suppliers_count_before | b4_suppliers_count_after
                if b4_suppliers_count_all:
                    b4_suppliers_count_new = b4_suppliers_count_after - b4_suppliers_count_before
                    b4_suppliers_count[buyer] = len(b4_suppliers_count_new) / len(b4_suppliers_count_all)

                    b5_suppliers_amount_all = 0
                    b5_suppliers_amount_new = 0

                    for compiled_release in compiled_releases:
                        ocid = compiled_release['ocid']
                        awards = compiled_release['awards']

                        for award in awards:
                            amount = award['value']['amount']
                            currency = award['value']['currency']

                            # Ensure we sum a single currency.
                            if not self.args.currency or currency == self.args.currency:
                                b5_suppliers_amount_all += amount
                                if any(supplier['identifier']['id'] in b4_suppliers_count_new for supplier in award['suppliers']):  # noqa
                                    b5_suppliers_amount_new += amount
                            else:
                                b5_other_currencies[str(currency)] += amount  # if null

                    b5_suppliers_amount[buyer] = b5_suppliers_amount_new / b5_suppliers_amount_all

                if b8_tenderers_count:
                    b8_tenderers_to_suppliers[buyer] = b8_tenderers_count / b8_suppliers_count

                if b1_tenderers_all:
                    b8_tenderers_never_suppliers[buyer] = (len(b1_tenderers_all - b4_suppliers_count_all) /
                                                           len(b1_tenderers_all))

        # Report indicators.

        def report_by_label(message, data, fmt='{1:-3.0%} {0}'):
            fmt = '  ' + fmt
            print(message)
            print(repr(data))
            for label, value in sorted(data.items()):
                print(fmt.format(label, value))
            print()

        def report_percentage_of_total(message, data, indicator, fmt='{}'):
            print(message)

            total = sum(data.values())
            data_format = '  {}: {:.0%} (' + fmt + ')'
            for label, count in sorted(data.items()):
                print(data_format.format(label, count / total, count))
            print()

        report_by_label('B0. % of new compiled releases:', b0_compiled_releases)
        report_by_label('B1. % of new bidders to all bidders:', b1_tenderers)
        report_percentage_of_total('B2. % of contracts assigned under each of the awarding procedures:',
                                   b2_procurement_method_count, 'B2')
        report_percentage_of_total('B3. % of total tender value assigned under each of the awarding procedures:',
                                   b3_procurement_method_amount, 'B3', fmt='${:,.2f}')
        print('B3. No data: {} (${:,.2f})'.format(len(b3_no_data), sum(b3_no_data)))
        report_by_label('B4. % of new winners to all winners:', b4_suppliers_count)
        report_by_label('B5. % of total contracting value awarded to new winners:', b5_suppliers_amount)
        report_by_label('B5.1. Excluded currencies:', b5_other_currencies, fmt='{0}: ${1:,.2f}')
        report_by_label('B8. Average number of bids to win:', b8_tenderers_to_suppliers, fmt='{1:,.2f} {0}')
        report_by_label('B8.1. % of bidders who never win:', b8_tenderers_never_suppliers)

        print('B9. Median number of bidders per tender:')
        for label, values in b9_number_of_tenderers.items():
            print('  {}: {}'.format(label, median(values)))
        print('B9. No data: {}\n'.format(b9_no_data))



import collections
import json

import ocdsmerge

from .base import BaseCommand


class Command(BaseCommand):
    name = 'compile'
    help = 'reads release packages from standard input, merges the releases by OCID, and prints the compiled releases'

    def handle(self):
        releases_by_ocid = collections.defaultdict(list)

        for line in self.buffer():
            release_package = json.loads(line)
            for release in release_package['releases']:
                releases_by_ocid[release['ocid']].append(release)

        for releases in releases_by_ocid.values():
            compiled_release = ocdsmerge.merge(releases)
            self.print(compiled_release)
